@@ -1,12 +1,31 @@
 #!/usr/bin/python
 # -- Content-Encoding: UTF-8 --
+"""
+The configuration module.
+"""
+
+__version__ = "0.1.5"
+
+# ------------------------------------------------------------------------------
 
 import sys
 
+# ------------------------------------------------------------------------------
+
 class LocalClasses(dict):
+    """
+    Associates local classes with their names (used in the jsonclass module)
+    """
     def add(self, cls):
+        """
+        Stores a local class
+        
+        :param cls: A class
+        """
         self[cls.__name__] = cls
 
+# ------------------------------------------------------------------------------
+
 class Config(object):
     """
     This is pretty much used exclusively for the 'jsonclass' 
@@ -14,28 +33,45 @@ class Config(object):
     You can change serialize_method and ignore_attribute, or use
     the local_classes.add(class) to include "local" classes.
     """
-    use_jsonclass = True
     # Change to False to keep __jsonclass__ entries raw.
-    serialize_method = '_serialize'
+    use_jsonclass = True
+
     # The serialize_method should be a string that references the
     # method on a custom class object which is responsible for
     # returning a tuple of the constructor arguments and a dict of
     # attributes.
-    ignore_attribute = '_ignore'
+    serialize_method = '_serialize'
+
     # The ignore attribute should be a string that references the
     # attribute on a custom class object which holds strings and / or
     # references of the attributes the class translator should ignore.
-    classes = LocalClasses()
+    ignore_attribute = '_ignore'
+
     # The list of classes to use for jsonclass translation.
+    classes = LocalClasses()
+
+    # Version of the JSON-RPC specification to support
     version = 2.0
-    # Version of the JSON-RPC spec to support
-    user_agent = 'jsonrpclib/0.1 (Python %s)' % \
-        '.'.join([str(ver) for ver in sys.version_info[0:3]])
+
     # User agent to use for calls.
+    user_agent = 'jsonrpclib-pelix/{0} (Python {1})' \
+                 .format(__version__,
+                         '.'.join(str(ver) for ver in sys.version_info[0:3]))
+
+    # Content-type to use. According to the JSON-RPC specification,
+    # it SHOULD be 'application/json-rpc'
+    # but MAY be 'application/json' or 'application/jsonrequest'
+    content_type = "application/json-rpc"
+
+    # "Singleton" of Config
     _instance = None
 
     @classmethod
     def instance(cls):
+        """
+        Returns/Creates the instance of Config
+        """
         if not cls._instance:
             cls._instance = cls()
+
         return cls._instance
