@@ -48,10 +48,16 @@ appropriately.
 See http://code.google.com/p/jsonrpclib/ for more info.
 """
 
+__version__ = "0.1.5"
+
+# ------------------------------------------------------------------------------
+
 # Library includes
-from jsonrpclib import config, history, utils
+from jsonrpclib import config
+import jsonrpclib.utils as utils
 
 # Standard library
+import contextlib
 import sys
 import uuid
 
@@ -75,6 +81,10 @@ else:
 
 # ------------------------------------------------------------------------------
 # JSON library import
+
+# JSON class serialization
+from jsonrpclib import jsonclass
+
 try:
     # Using cjson
     import cjson
@@ -116,20 +126,102 @@ except ImportError:
 # XMLRPClib re-implementations
 
 class ProtocolError(Exception):
+    """
+    JSON-RPC error
+    
+    ProtocolError[0] can be:
+    * an error message (string)
+    * a (code, message) tuple
+    """
     pass
 
+class AppError(ProtocolError):
+    """
+    Application error: the error code is not in the pre-defined ones
+    
+    AppError[0][0]: Error code
+    AppError[0][1]: Error message or trace
+    AppError[0][2]: Associated data
+    """
+    def data(self):
+        """
+        Retrieves the value found in the 'data' entry of the error, or None
+        
+        :return: The data associated to the error, or None
+        """
+        return self[0][2]
+
+
 class TransportMixIn(object):
     """ Just extends the XMLRPC transport where necessary. """
     user_agent = config.user_agent
     # for Python 2.7 support
     _connection = None
 
+    # Additional headers: list of dictionaries
+    additional_headers = []
+
+    # List of non-overridable headers
+    # Use the configuration to change the content-type
+    readonly_headers = ('content-length', 'content-type')
+
+    def push_headers(self, headers):
+        """
+        Adds a dictionary of headers to the additional headers list
+        
+        :param headers: A dictionary
+        """
+        self.additional_headers.append(headers)
+
+    def pop_headers(self, headers):
+        """
+        Removes the given dictionary from the additional headers list. 
+        Also validates that given headers are on top of the stack
+        
+        :param headers: Headers to remove
+        :raise AssertionError: The given dictionary is not on the latest stored
+                               in the additional headers list
+        """
+        assert self.additional_headers[-1] == headers
+        self.additional_headers.pop()
+
+
+    def emit_additional_headers(self, connection):
+        """
+        Puts headers as is in the request, filtered read only headers
+        
+        :param connection: The request connection
+        """
+        additional_headers = {}
+
+        # Prepare the merged dictionary
+        for headers in self.additional_headers:
+            additional_headers.update(headers)
+
+        # Remove forbidden keys
+        for forbidden in self.readonly_headers:
+            additional_headers.pop(forbidden, None)
+
+        # Reversed order: in the case of multiple headers value definition,
+        # the latest pushed has priority
+        for key, value in additional_headers.items():
+            key = str(key)
+            if key.lower() not in self.readonly_headers:
+                # Only accept replaceable headers
+                connection.putheader(str(key), str(value))
+
+
     def send_content(self, connection, request_body):
         # Convert the body first
         request_body = utils.to_bytes(request_body)
 
-        connection.putheader("Content-Type", "application/json-rpc")
+        # "static" headers
+        connection.putheader("Content-Type", config.content_type)
         connection.putheader("Content-Length", str(len(request_body)))
+
+        # Emit additional headers here in order not to override content-length
+        self.emit_additional_headers(connection)
+
         connection.endheaders()
         if request_body:
             connection.send(request_body)
@@ -153,15 +245,23 @@ class JSONTarget(object):
         self.data = []
 
     def feed(self, data):
-        try:
-            data = utils.from_bytes(data)
-        except:
-            pass
-
+        # Store raw data: it might not contain whole wide-character
         self.data.append(data)
 
     def close(self):
-        return ''.join(self.data)
+        if not self.data:
+            return ''
+
+        else:
+            data = type(self.data[0])().join(self.data)
+            try:
+                # Convert the whole final string
+                data = utils.from_bytes(data)
+            except:
+                # Try a pass-through
+                pass
+
+            return data
 
 class Transport(TransportMixIn, XMLTransport):
     pass
@@ -178,10 +278,22 @@ class ServerProxy(XMLServerProxy):
     """
 
     def __init__(self, uri, transport=None, encoding=None,
-                 verbose=0, version=None):
+                 verbose=0, version=None, headers=None, history=None):
+        """
+        Sets up the server proxy
+        
+        :param uri: Request URI
+        :param transport: Custom transport handler
+        :param encoding: Specified encoding
+        :param verbose: Log verbosity level
+        :param version: JSON-RPC specification version
+        :param headers: Custom additional headers for each request
+        :param history: History object (for tests)
+        """
         if not version:
             version = config.version
         self.__version = version
+
         schema, uri = splittype(uri)
         if schema not in ('http', 'https'):
             raise IOError('Unsupported JSON-RPC protocol.')
@@ -197,8 +309,13 @@ class ServerProxy(XMLServerProxy):
             else:
                 transport = Transport()
         self.__transport = transport
+
         self.__encoding = encoding
         self.__verbose = verbose
+        self.__history = history
+
+        # Global custom headers are injected into Transport
+        self.__transport.push_headers(headers or {})
 
     def _request(self, methodname, params, rpcid=None):
         request = dumps(params, methodname, encoding=self.__encoding,
@@ -215,7 +332,8 @@ class ServerProxy(XMLServerProxy):
         return
 
     def _run_request(self, request, notify=None):
-        history.add_request(request)
+        if self.__history is not None:
+            self.__history.add_request(request)
 
         response = self.__transport.request(
             self.__host,
@@ -230,7 +348,9 @@ class ServerProxy(XMLServerProxy):
         # the response object, or expect the Server to be
         # outputting the response appropriately?
 
-        history.add_response(response)
+        if self.__history is not None:
+            self.__history.add_response(response)
+
         if not response:
             return None
         return_obj = loads(response)
@@ -245,6 +365,21 @@ class ServerProxy(XMLServerProxy):
         # Just like __getattr__, but with notify namespace.
         return _Notify(self._request_notify)
 
+    @contextlib.contextmanager
+    def _additional_headers(self, headers):
+        """
+            Allow to specify additional headers, to be added inside the with
+            block. Example of usage:
+
+            >>> with client._additional_headers({'X-Test' : 'Test'}) as new_client:
+            ...     new_client.method()
+            ...
+            >>> # Here old headers are restored
+        """
+        self.__transport.push_headers(headers)
+        yield self
+        self.__transport.pop_headers(headers)
+
 # ------------------------------------------------------------------------------
 
 class _Method(XML_Method):
@@ -370,36 +505,96 @@ Server = ServerProxy
 # ------------------------------------------------------------------------------
 
 class Fault(object):
-    # JSON-RPC error class
-    def __init__(self, code= -32000, message='Server error', rpcid=None):
+    """
+    JSON-RPC error class
+    """
+    def __init__(self, code=-32000, message='Server error', rpcid=None):
+        """
+        Sets up the error description
+        
+        :param code: Fault code
+        :param message: Associated message
+        :param rpcid: Request ID
+        """
         self.faultCode = code
         self.faultString = message
         self.rpcid = rpcid
 
     def error(self):
+        """
+        Returns the error as a dictionary
+        
+        :returns: A {'code', 'message'} dictionary
+        """
         return {'code':self.faultCode, 'message':self.faultString}
 
     def response(self, rpcid=None, version=None):
+        """
+        Returns the error as a JSON-RPC response string
+        
+        :param rpcid: Forced request ID
+        :param version: JSON-RPC version
+        :return: A JSON-RPC response string
+        """
         if not version:
             version = config.version
+
         if rpcid:
             self.rpcid = rpcid
-        return dumps(
-            self, methodresponse=True, rpcid=self.rpcid, version=version
-        )
 
-    def __repr__(self):
-        return '<Fault %s: %s>' % (self.faultCode, self.faultString)
+        return dumps(self, methodresponse=True, rpcid=self.rpcid,
+                     version=version)
 
-class Payload(dict):
-    def __init__(self, rpcid=None, version=None):
-        dict.__init__(self)
+    def dump(self, rpcid=None, version=None):
+        """
+        Returns the error as a JSON-RPC response dictionary
+        
+        :param rpcid: Forced request ID
+        :param version: JSON-RPC version
+        :return: A JSON-RPC response dictionary
+        """
         if not version:
             version = config.version
+
+        if rpcid:
+            self.rpcid = rpcid
+
+        return dump(self, is_response=True, rpcid=self.rpcid,
+                    version=version)
+
+    def __repr__(self):
+        """
+        String representation
+        """
+        return '<Fault {0}: {1}>'.format(self.faultCode, self.faultString)
+
+
+class Payload(object):
+    """
+    JSON-RPC content handler
+    """
+    def __init__(self, rpcid=None, version=None):
+        """
+        Sets up the JSON-RPC handler
+        
+        :param rpcid: Request ID
+        :param version: JSON-RPC version
+        """
+        if not version:
+            version = config.version
+
         self.id = rpcid
         self.version = float(version)
 
+
     def request(self, method, params=[]):
+        """
+        Prepares a method call request
+        
+        :param method: Method name
+        :param params: Method parameters
+        :return: A JSON-RPC request dictionary
+        """
         if type(method) not in utils.StringTypes:
             raise ValueError('Method name must be a string.')
 
@@ -408,29 +603,60 @@ class Payload(dict):
             self.id = str(uuid.uuid4())
 
         request = { 'id':self.id, 'method':method }
-        if params:
+        if params or self.version < 1.1:
             request['params'] = params
+
         if self.version >= 2:
             request['jsonrpc'] = str(self.version)
+
         return request
 
+
     def notify(self, method, params=[]):
+        """
+        Prepares a notification request
+        
+        :param method: Notification name
+        :param params: Notification parameters
+        :return: A JSON-RPC notification dictionary
+        """
+        # Prepare the request dictionary
         request = self.request(method, params)
+
+        # Remove the request ID, as it's a notification
         if self.version >= 2:
             del request['id']
         else:
             request['id'] = None
+
         return request
 
+
     def response(self, result=None):
+        """
+        Prepares a response dictionary
+        
+        :param result: The result of method call
+        :return: A JSON-RPC response dictionary
+        """
         response = {'result':result, 'id':self.id}
+
         if self.version >= 2:
             response['jsonrpc'] = str(self.version)
         else:
             response['error'] = None
+
         return response
 
-    def error(self, code= -32000, message='Server error.'):
+
+    def error(self, code=-32000, message='Server error.'):
+        """
+        Prepares an error dictionary
+        
+        :param code: Error code
+        :param message: Error message
+        :return: A JSON-RPC error dictionary
+        """
         error = self.response()
         if self.version >= 2:
             del error['result']
@@ -439,83 +665,196 @@ class Payload(dict):
         error['error'] = {'code':code, 'message':message}
         return error
 
-def dumps(params=[], methodname=None, methodresponse=None,
-          encoding=None, rpcid=None, version=None, notify=None):
+# ------------------------------------------------------------------------------
+
+def dump(params=[], methodname=None, rpcid=None, version=None,
+         is_response=None, is_notify=None):
     """
-    This differs from the Python implementation in that it implements
-    the rpcid argument since the 2.0 spec requires it for responses.
+    Prepares a JSON-RPC dictionary (request, notification, response or error)
+    
+    :param params: Method parameters (if a method name is given) or a Fault
+    :param methodname: Method name
+    :param rpcid: Request ID
+    :param version: JSON-RPC version
+    :param is_response: If True, this is a response dictionary
+    :param is_notify: If True, this is a notification request
+    :return: A JSON-RPC dictionary
     """
+    # Default version
     if not version:
         version = config.version
-    valid_params = (utils.TupleType, utils.ListType, utils.DictType)
+
+    # Validate method name and parameters
+    valid_params = (utils.TupleType, utils.ListType, utils.DictType, Fault)
     if methodname in utils.StringTypes and \
-            type(params) not in valid_params and \
-            not isinstance(params, Fault):
+    not isinstance(params, valid_params):
         """
         If a method, and params are not in a listish or a Fault,
         error out.
         """
-        raise TypeError('Params must be a dict, list, tuple or Fault ' +
-                        'instance.')
-    # Begin parsing object
+        raise TypeError('Params must be a dict, list, tuple or Fault instance.')
+
+    # Prepares the JSON-RPC content
     payload = Payload(rpcid=rpcid, version=version)
-    if not encoding:
-        encoding = 'utf-8'
+
     if type(params) is Fault:
-        response = payload.error(params.faultCode, params.faultString)
-        return jdumps(response, encoding=encoding)
-    if type(methodname) not in utils.StringTypes and methodresponse != True:
-        raise ValueError('Method name must be a string, or methodresponse ' +
+        # Prepare an error dictionary
+        return payload.error(params.faultCode, params.faultString)
+
+    if type(methodname) not in utils.StringTypes and not is_response:
+        # Neither a request nor a response
+        raise ValueError('Method name must be a string, or is_response ' \
                          'must be set to True.')
-    if config.use_jsonclass == True:
-        from jsonrpclib import jsonclass
+
+    if config.use_jsonclass:
+        # Use jsonclass to convert the parameters
         params = jsonclass.dump(params)
-    if methodresponse is True:
+
+    if is_response:
+        # Prepare a response dictionary
         if rpcid is None:
+            # A response must have a request ID
             raise ValueError('A method response must have an rpcid.')
-        response = payload.response(params)
-        return jdumps(response, encoding=encoding)
-    request = None
-    if notify == True:
-        request = payload.notify(methodname, params)
+        return payload.response(params)
+
+    if is_notify:
+        # Prepare a notification dictionary
+        return payload.notify(methodname, params)
+
     else:
-        request = payload.request(methodname, params)
+        # Prepare a method call dictionary
+        return payload.request(methodname, params)
+
+
+def dumps(params=[], methodname=None, methodresponse=None,
+          encoding=None, rpcid=None, version=None, notify=None):
+    """
+    Prepares a JSON-RPC request/response string
+    
+    :param params: Method parameters (if a method name is given) or a Fault
+    :param methodname: Method name
+    :param methodresponse: If True, this is a response dictionary
+    :param encoding: Result string encoding
+    :param rpcid: Request ID
+    :param version: JSON-RPC version
+    :param notify: If True, this is a notification request
+    :return: A JSON-RPC dictionary
+    """
+    # Prepare the dictionary
+    request = dump(params, methodname, rpcid, version, methodresponse, notify)
+
+    # Set the default encoding
+    if not encoding:
+        encoding = "UTF-8"
+
+    # Returns it as a JSON string
     return jdumps(request, encoding=encoding)
 
-def loads(data):
+
+def load(data):
     """
-    This differs from the Python implementation, in that it returns
-    the request structure in Dict format instead of the method, params.
-    It will return a list in the case of a batch request / response.
+    Loads a JSON-RPC request/response dictionary. Calls jsonclass to load beans
+    
+    :param data: A JSON-RPC dictionary
+    :return: A parsed dictionary or None
     """
-    if data == '':
-        # notification
+    if data is None:
+        # Notification
         return None
-    result = jloads(data)
+
     # if the above raises an error, the implementing server code
     # should return something like the following:
     # { 'jsonrpc':'2.0', 'error': fault.error(), id: None }
-    if config.use_jsonclass == True:
-        from jsonrpclib import jsonclass
-        result = jsonclass.load(result)
-    return result
+    if config.use_jsonclass:
+        # Convert beans
+        data = jsonclass.load(data)
+
+    return data
+
+
+def loads(data):
+    """
+    Loads a JSON-RPC request/response string. Calls jsonclass to load beans
+    
+    :param data: A JSON-RPC string
+    :return: A parsed dictionary or None
+    """
+    if data == '':
+        # Notification
+        return None
+
+    # Parse the JSON dictionary
+    result = jloads(data)
+
+    # Load the beans
+    return load(result)
+
+# ------------------------------------------------------------------------------
 
 def check_for_errors(result):
+    """
+    Checks if a result dictionary signals an error
+    
+    :param result: A result dictionary
+    :raise TypeError: Invalid parameter
+    :raise NotImplementedError: Unknown JSON-RPC version
+    :raise ValueError: Invalid dictionary content
+    :raise ProtocolError: An error occurred on the server side
+    :return: The result parameter
+    """
     if not result:
         # Notification
         return result
+
     if type(result) is not utils.DictType:
+        # Invalid argument
         raise TypeError('Response is not a dict.')
-    if 'jsonrpc' in result.keys() and float(result['jsonrpc']) > 2.0:
+
+    if 'jsonrpc' in result and float(result['jsonrpc']) > 2.0:
+        # Unknown JSON-RPC version
         raise NotImplementedError('JSON-RPC version not yet supported.')
-    if 'result' not in result.keys() and 'error' not in result.keys():
+
+    if 'result' not in result and 'error' not in result:
+        # Invalid dictionary content
         raise ValueError('Response does not have a result or error key.')
-    if 'error' in result.keys() and result['error'] != None:
-        code = result['error']['code']
-        message = result['error']['message']
-        raise ProtocolError((code, message))
+
+    if 'error' in result and result['error']:
+        # Server-side error
+        if 'code' in result['error']:
+            # Code + Message
+            code = result['error']['code']
+            try:
+                # Get the message (jsonrpclib)
+                message = result['error']['message']
+
+            except KeyError:
+                # Get the trace (jabsorb)
+                message = result['error'].get('trace', '<no error message>')
+
+            if -32700 <= code <= -32000:
+                # Pre-defined errors
+                # See http://www.jsonrpc.org/specification#error_object
+                raise ProtocolError((code, message))
+
+            else:
+                # Application error
+                data = result['error'].get('data', None)
+                raise AppError((code, message, data))
+
+            raise ProtocolError((code, message))
+
+        elif isinstance(result['error'], dict) and len(result['error']) == 1:
+            # Error with a single entry ('reason', ...): use its content
+            error_key = result['error'].keys()[0]
+            raise ProtocolError(result['error'][error_key])
+
+        else:
+            # Use the raw error content
+            raise ProtocolError(result['error'])
+
     return result
 
+
 def isbatch(result):
     if type(result) not in (utils.ListType, utils.TupleType):
         return False
@@ -533,11 +872,20 @@ def isbatch(result):
         return False
     return True
 
+
 def isnotification(request):
-    if 'id' not in request.keys():
+    """
+    Tests if the given request is a notification
+    
+    :param request: A request dictionary
+    :return: True if the request is a notification
+    """
+    if 'id' not in request:
         # 2.0 notification
         return True
-    if request['id'] == None:
+
+    if request['id'] is None:
         # 1.0 notification
         return True
+
     return False
